from typing import TypeVar, Any, Awaitable, Callable, Type

from paypal2.models.hook import (
    WebHookEvent,
    WebHookEventCaptureCompleted,
    WebHookEventCaptureReversed,
    WebHookEventCaptureRefunded,
    WebHookEventSubscriptionCreated,
    WebHookEventSubscriptionActivated,
    WebHookEventSubscriptionSuspended,
    WebHookEventSubscriptionExpired,
    WebHookEventSubscriptionCancelled,
    WebHookEventSubscriptionPaymentFailed,
    WebHookEventSaleCompleted,
    WebHookEventSaleReversed,
    WebHookEventSaleRefunded,
    WebHookEventPlanCreated,
    WebHookEventPlanUpdated,
)

_WebHookEventType = TypeVar("_WebHookEventType", bound=WebHookEvent)
WebHookHandler = Callable[[_WebHookEventType], Awaitable[Any]]
WebHookHandlers = dict[Type[_WebHookEventType], WebHookHandler]


class PayPalWebHookProcessorBase:
    """
    WebHook processor class for processing the raw events freely.
    """

    __slots__ = ("webhook_handlers",)

    def __init__(self):
        self.webhook_handlers = {e.HANDLER_KEY: (e, h) for e, h in self._init_handlers().items()}

    def _init_handlers(self) -> WebHookHandlers:
        """
        Override this method to remove or extend the used handlers.
        """
        return {
            # WebHookEventCapturePending: self.event_capture_pending,
            WebHookEventCaptureCompleted: self.event_capture_completed,
            WebHookEventCaptureReversed: self.event_capture_reversed,
            WebHookEventCaptureRefunded: self.event_capture_refunded,
            # WebHookEventCaptureDeclined: self.event_capture_declined,
            WebHookEventSubscriptionCreated: self.event_subscription_created,
            WebHookEventSubscriptionActivated: self.event_subscription_activated,
            WebHookEventSubscriptionSuspended: self.event_subscription_suspended,
            WebHookEventSubscriptionExpired: self.event_subscription_expired,
            WebHookEventSubscriptionCancelled: self.event_subscription_cancelled,
            WebHookEventSubscriptionPaymentFailed: self.event_subscription_payment_failed,
            WebHookEventSaleCompleted: self.event_sale_completed,
            WebHookEventSaleReversed: self.event_sale_reversed,
            WebHookEventSaleRefunded: self.event_sale_refunded,
            WebHookEventPlanCreated: self.event_plan_created,
            WebHookEventPlanUpdated: self.event_plan_updated,
            WebHookEvent: self.event_fallback,
        }

    async def call(self, data: dict) -> Any:
        """
        Automatic WebHook data handler selection and processing.
        """
        handler_key = (data.get("event_type"), data.get("resource_type"), data.get("resource_version", "1.0"))
        event_handler = self.webhook_handlers.get(handler_key)
        if not event_handler:
            event_handler = self.webhook_handlers.get(None)  # try fallback if it is registered
            if not event_handler:
                return
        event = event_handler[0].model_validate(data)
        return await event_handler[1](event)

    async def event_capture_completed(self, event: WebHookEventCaptureCompleted) -> Any:
        """
        This is mostly informational, as the capturing of the order (unless you use intent: AUTHORIZE) is
        done by direct API call.
        """

    async def event_capture_reversed(self, event: WebHookEventCaptureReversed) -> Any:
        """
        The server shall revoke the granted resource by the id of the capture or record it as denied for reference.
        """

    async def event_capture_refunded(self, event: WebHookEventCaptureRefunded) -> Any:
        """
        The server shall revoke the granted resource by the id of the capture or record it as denied for reference.
        """

    async def event_subscription_created(self, event: WebHookEventSubscriptionCreated) -> Any:
        """
        This is the first event where the new subscription-id (id), plan_id and custom_id are all present and have
        clear relations to each other. The server should record these associations for later, for example a
        sale-completed event will only connect to the subscription through the custom_id (queried from the captured
        payment endpoint).
        Tracking the subscription for a user/identity may be desired for information or performing actions like
        suspend or continue for example anyway.
        """

    async def event_subscription_activated(self, event: WebHookEventSubscriptionActivated) -> Any:
        """
        Record the subscription state change for the relevant party if it is tracked.
        """

    async def event_subscription_suspended(self, event: WebHookEventSubscriptionSuspended) -> Any:
        """
        Record the subscription state change for the relevant party if it is tracked.
        """

    async def event_subscription_expired(self, event: WebHookEventSubscriptionExpired) -> Any:
        """
        Record the subscription state change or remove the subscription from the relevant party if it is tracked,
        because it is no longer valid.
        """

    async def event_subscription_cancelled(self, event: WebHookEventSubscriptionCancelled) -> Any:
        """
        Record the subscription state change or remove the subscription from the relevant party if it is tracked,
        because it is no longer valid.
        """

    async def event_subscription_payment_failed(self, event: WebHookEventSubscriptionPaymentFailed) -> Any:
        """
        A payment was not successfully charged. This is mostly informational, the relevant party should be notified.
        The subscription may still be active or may have got suspended based on the payment_preferences of the plan.
        NOTE: It is not clear if failure of charging the setup_fee would trigger this event, but depending on
        the payment_preferences, the subscription could also be cancelled here.
        """

    async def event_sale_completed(self, event: WebHookEventSaleCompleted) -> Any:
        """
        A subscription payment has been successfully charged, grant the service/resource to the relevant party.
        """

    async def event_sale_reversed(self, event: WebHookEventSaleReversed) -> Any:
        """
        The server shall revoke the granted resource by the id of the sale or record it as denied for reference.
        """

    async def event_sale_refunded(self, event: WebHookEventSaleRefunded) -> Any:
        """
        The server shall revoke the granted resource by the id of the sale or record it as denied for reference.
        """

    async def event_plan_created(self, event: WebHookEventPlanCreated) -> Any:
        """
        The server can track the details of specific plans that are relevant for its users. This is beneficial,
        as the products and plans are primarily managed on the PayPal server/UI.
        """

    async def event_plan_updated(self, event: WebHookEventPlanUpdated) -> Any:
        """
        The server can track the details of specific plans that are relevant for its users. This is beneficial,
        as the products and plans are primarily managed on the PayPal server/UI.
        """

    async def event_fallback(self, event: WebHookEvent) -> Any:
        """
        Any event not captured by specific handlers may be processed here for introspection for example.
        """



from typing import Optional

from paypal2.models.common import MonetaryValue, PlanBillingCycle
from paypal2.models.order import OrdersCreate, PurchaseUnit, Item, MonetaryValueWithBreakdown, Breakdown


def get_regular_billing_cycle(billing_cycles: Optional[list[PlanBillingCycle]]) -> Optional[PlanBillingCycle]:
    for bc in billing_cycles or []:
        if bc.tenure_type == "REGULAR":
            if bc.pricing_scheme and bc.frequency:
                return bc


def format_monetary_value(v: float | int) -> str:
    if isinstance(v, float):
        return "%.2f" % v
    return str(v)


def simple_single_item_order_create(
    item_name: str,
    price: float | int,
    currency_code: str = "USD",
    custom_id: Optional[str] = None,
    invoice_id: Optional[str] = None,
    tax: Optional[float | int] = None,
    quantity: int = 1,
    category: str = "DIGITAL_GOODS",
    intent: str = "CAPTURE",
    reference_id: Optional[str] = None,  # NOTE: does not appear in webhook messages
) -> OrdersCreate:
    assert quantity > 0
    unit_amount = MonetaryValue(currency_code=currency_code, value=format_monetary_value(price))
    total_price = price * quantity
    item_total = MonetaryValue(currency_code=currency_code, value=format_monetary_value(total_price))
    total = total_price
    if tax is None:
        tax_total = unit_tax = None
    else:
        unit_tax = MonetaryValue(currency_code=currency_code, value=format_monetary_value(tax))
        total_tax = tax * quantity
        tax_total = MonetaryValue(currency_code=currency_code, value=format_monetary_value(total_tax))
        total += total_tax
    return OrdersCreate(
        purchase_units=[
            PurchaseUnit(
                reference_id=reference_id,
                custom_id=custom_id,
                invoice_id=invoice_id,
                items=[
                    Item(
                        name=item_name,
                        quantity=str(quantity),
                        category=category,
                        unit_amount=unit_amount,
                        tax=unit_tax,
                    ),
                ],
                amount=MonetaryValueWithBreakdown(
                    currency_code=currency_code,
                    value=format_monetary_value(total),
                    breakdown=Breakdown(
                        item_total=item_total,
                        tax_total=tax_total,
                    ),
                ),
            )
        ],
        intent=intent,
    )



from typing import Optional

import aiohttp

from paypal2.client import PayPalApiClient
from paypal2.models.common import (
    PlanBillingCycle,
    MonetaryValue,
    PlanPricingModel,
    PlanFrequency,
    PlanPaymentPreferences,
    PlanTaxes,
)
from paypal2.models.plan import PlanDetails, PlanCreate
from paypal2.models.product import ProductCreate


async def ensure_simple_product_plan(
    pp: PayPalApiClient,
    product_id: str,
    plan_name: str,
    product_name: str = "digital subscription plan for testing",
    plan_fixed_price=MonetaryValue(currency_code="USD", value="6"),
    plan_frequency=PlanFrequency(interval_unit="MONTH", interval_count=1),
    plan_setup_fee=MonetaryValue(currency_code="USD", value="0"),
    plan_taxes: Optional[PlanTaxes] = None,
) -> Optional[PlanDetails]:
    """
    Subscription plans should be managed on the PayPal business UI usually.
    This code is mostly provided here for conveniently creating a specific plan-setup during testing.
    """
    if plan_fixed_price.currency_code != plan_setup_fee.currency_code:
        if float(plan_setup_fee.value) == 0:
            plan_setup_fee = MonetaryValue(currency_code=plan_fixed_price.currency_code, value="0")
        else:
            raise ValueError("currency_code must match for price and setup fee")
    try:
        plan_list = await pp.plan_list(product_id, page_size=20)
        for plan in plan_list.plans:
            if plan.name == plan_name and plan.product_id == product_id:
                break
        else:
            plan = None
    except aiohttp.ClientResponseError as e:
        if e.status != 404:
            raise
        plan = None
    if plan is None:
        product = await pp.product_details(product_id)
        if product is None:
            await pp.product_create(ProductCreate(id=product_id, name=product_name, type="DIGITAL"))
        plan = await pp.plan_create(
            PlanCreate(
                product_id=product_id,
                name=plan_name,
                billing_cycles=[
                    PlanBillingCycle(
                        tenure_type="REGULAR",
                        sequence=1,
                        total_cycles=0,
                        pricing_scheme=PlanPricingModel(fixed_price=plan_fixed_price),
                        frequency=plan_frequency,
                    ),
                ],
                payment_preferences=PlanPaymentPreferences(setup_fee=plan_setup_fee),
                taxes=plan_taxes,
            )
        )
    plan_id = plan.id
    return await pp.plan_details(plan_id)



from typing import Any, Optional

from paypal2.client import PayPalApiClient
from paypal2.hook import PayPalWebHookProcessorBase
from paypal2.models.common import PlanBillingCycle
from paypal2.models.hook import (
    WebHookEventCaptureReversed,
    WebHookEventCaptureRefunded,
    WebHookEventSubscriptionCreated,
    WebHookEventSubscriptionExpired,
    WebHookEventSubscriptionCancelled,
    WebHookEventSubscriptionPaymentFailed,
    WebHookEventSaleCompleted,
    WebHookEventSaleReversed,
    WebHookEventSaleRefunded,
    WebHookEventPlanCreated,
    WebHookEventPlanUpdated,
    WebHookSaleResource,
    WebHookSubscriptionResourceV2,
    WebHookEventSubscriptionActivated,
    WebHookEventSubscriptionSuspended,
)
from paypal2.models.payment import CapturedPayment
from paypal2.utility import get_regular_billing_cycle


class AbstractPayPalWebHookProcessor(PayPalWebHookProcessorBase):
    """
    WebHook processor class with some general pre-processing and improved handler logic.

    The extended logic relies on the custom_id being used for subscriptions and the server being able to decode the
    identity of the subscribing entity and the service being purchased.

    NOTE: This can't be used with simulated webhook messages, as they do not provide real ids for external entities.
    """

    __slots__ = ("pp",)

    def __init__(self, pp: PayPalApiClient):
        PayPalWebHookProcessorBase.__init__(self)
        self.pp = pp

    async def _get_payment_details_with_custom_id(
        self, resource: WebHookSubscriptionResourceV2 | WebHookSaleResource
    ) -> Optional[CapturedPayment]:
        captured_payment = await self.pp.captured_payment_details(resource.id)
        if captured_payment and captured_payment.custom_id:
            return captured_payment

    # async def event_capture_completed(self, event: WebHookEventCaptureCompleted) -> Any:
    #     pass

    async def event_capture_reversed(self, event: WebHookEventCaptureReversed) -> Any:
        return await self._revoke_order(event)

    async def event_capture_refunded(self, event: WebHookEventCaptureRefunded) -> Any:
        return await self._revoke_order(event)

    async def event_subscription_created(self, event: WebHookEventSubscriptionCreated) -> Any:
        if event.resource.custom_id:
            return await self._associate_subscription(event)

    async def event_subscription_activated(self, event: WebHookEventSubscriptionActivated) -> Any:
        if event.resource.custom_id:
            return await self._track_subscription(event)

    async def event_subscription_suspended(self, event: WebHookEventSubscriptionSuspended) -> Any:
        if event.resource.custom_id:
            return await self._track_subscription(event)

    async def event_subscription_expired(self, event: WebHookEventSubscriptionExpired) -> Any:
        if event.resource.custom_id:
            return await self._dissociate_subscription(event)

    async def event_subscription_cancelled(self, event: WebHookEventSubscriptionCancelled) -> Any:
        if event.resource.custom_id:
            return await self._dissociate_subscription(event)

    async def event_subscription_payment_failed(self, event: WebHookEventSubscriptionPaymentFailed) -> Any:
        if event.resource.custom_id:
            return await self._warn_subscription_failure(event)

    async def event_sale_completed(self, event: WebHookEventSaleCompleted) -> Any:
        payment_details = await self._get_payment_details_with_custom_id(event.resource)
        if payment_details and payment_details.status == "COMPLETED":
            # NOTE: Not sure if status can be anything but "COMPLETED", or what should be done then.
            return await self._grant_subscription(event, payment_details)

    async def event_sale_reversed(self, event: WebHookEventSaleReversed) -> Any:
        payment_details = await self._get_payment_details_with_custom_id(event.resource)
        if payment_details:
            return await self._revoke_subscription(event, payment_details)

    async def event_sale_refunded(self, event: WebHookEventSaleRefunded) -> Any:
        payment_details = await self._get_payment_details_with_custom_id(event.resource)
        if payment_details:
            return await self._revoke_subscription(event, payment_details)

    async def event_plan_created(self, event: WebHookEventPlanCreated) -> Any:
        regular_bc = get_regular_billing_cycle(event.resource.billing_cycles)
        if regular_bc:
            return await self._track_plan_pricing(event, regular_bc)

    async def event_plan_updated(self, event: WebHookEventPlanUpdated) -> Any:
        regular_bc = get_regular_billing_cycle(event.resource.billing_cycles)
        if regular_bc:
            return await self._track_plan_pricing(event, regular_bc)

    # async def event_fallback(self, event: WebHookEvent) -> Any:
    #     pass

    async def _revoke_order(self, event: WebHookEventCaptureReversed | WebHookEventCaptureRefunded) -> Any:
        """
        Revoke the ordered service denoted by the id of the payment.
        """
        raise NotImplementedError()

    async def _associate_subscription(self, event: WebHookEventSubscriptionCreated) -> Any:
        """
        The relationship among subscription-id, plan-id, custom-id and likely user/party must be recorded so later
        "grant_subscription" events can look up all the information from a "custom_id".
        """
        raise NotImplementedError()

    async def _track_subscription(
        self, event: WebHookEventSubscriptionActivated | WebHookEventSubscriptionSuspended
    ) -> Any:
        """
        For tracking a subscription status on the server.
        """
        raise NotImplementedError()

    async def _dissociate_subscription(
        self, event: WebHookEventSubscriptionExpired | WebHookEventSubscriptionCancelled
    ) -> Any:
        """
        For tracking a subscription on the server, mark or discard the subscription resource-id from the subscriber
        denoted by the custom_id.
        """
        raise NotImplementedError()

    async def _warn_subscription_failure(self, event: WebHookEventSubscriptionPaymentFailed) -> Any:
        """
        Notify the subscriber denoted by the custom_id that their subscription payment could not be processed.
        """
        raise NotImplementedError()

    async def _grant_subscription(self, event: WebHookEventSaleCompleted, payment_details: CapturedPayment) -> Any:
        """
        Grant the subscribed service denoted by the custom_id since created_time, tracked as the id of the payment.
        """
        raise NotImplementedError()

    async def _revoke_subscription(
        self, event: WebHookEventSaleReversed | WebHookEventSaleRefunded, payment_details: CapturedPayment
    ) -> Any:
        """
        Revoke the subscribed service denoted by the custom_id and tracked as the id of the payment.
        """
        raise NotImplementedError()

    async def _track_plan_pricing(
        self, event: WebHookEventPlanCreated | WebHookEventPlanUpdated, billing_cycle: PlanBillingCycle
    ) -> Any:
        """
        Track the property changes of the plan resource. The "REGULAR" billing cycle is selected for convenience.
        NOTE: The server must query and update the tracked plans on startup using the plan_details(...) api and only
        rely on the webhooks to keep the state synced.
        """
        raise NotImplementedError()



from datetime import datetime
from typing import Optional, ClassVar

from pydantic import BaseModel, Field

from paypal2.models.common import (
    HATEOASLink,
    PlanBillingCycle,
    PlanPaymentPreferences,
    PlanTaxes,
    MonetaryValue,
    PaymentSupplementaryData,
)


class WebHookEvent(BaseModel):
    HANDLER_KEY: ClassVar[Optional[tuple[str, str, str]]] = None
    id: str
    event_type: str
    event_version: str
    resource_type: str
    resource_version: Optional[str] = None
    resource: dict
    links: list[HATEOASLink]


class WebHookCaptureResourceV2(BaseModel):
    id: str
    status: str
    amount: Optional[MonetaryValue] = None
    supplementary_data: Optional[PaymentSupplementaryData] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # links: list[HATEOASLink]
    # todo seller_protection
    # todo seller_receivable_breakdown
    # todo payee
    # todo final_capture


class WebHookRefundResourceV2(BaseModel):
    id: str
    status: str
    amount: Optional[MonetaryValue] = None
    supplementary_data: Optional[PaymentSupplementaryData] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # links: list[HATEOASLink]
    # todo seller_protection
    # todo seller_receivable_breakdown
    # todo final_capture
    # todo disbursement_mode
    # todo invoice_id
    # todo custom_id


class WebHookSubscriptionResourceV2(BaseModel):
    id: str
    status: str
    status_update_time: Optional[str] = None
    plan_id: str
    custom_id: Optional[str] = None
    start_time: Optional[str] = None
    auto_renewal: Optional[bool] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # links: list[HATEOASLink]
    # todo quantity
    # todo shipping_amount
    # todo subscriber
    # todo billing_info


class WebHookPlanResourceV2(BaseModel):
    id: str
    product_id: str
    status: str
    name: Optional[str] = None
    description: Optional[str] = None
    usage_type: Optional[str] = None
    billing_cycles: Optional[list[PlanBillingCycle]] = None
    payment_preferences: Optional[PlanPaymentPreferences] = None
    taxes: Optional[PlanTaxes] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # links: list[HATEOASLink]


class MonetaryTotal(BaseModel):
    currency: str = Field(max_length=5)
    total: str = Field(max_length=50)
    details: Optional[dict] = None


class WebHookSaleResource(BaseModel):
    id: str
    state: str
    amount: Optional[MonetaryTotal] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None  # for complete and refund
    # links: list[HATEOASLink]
    # todo sale_id for refund
    # todo parent_payment for complete, refund
    # todo clearing_time for complete
    # todo payment_mode for complete
    # todo protection_eligibility for complete
    # todo protection_eligibility_type for complete

    # NOTE: There is a "custom" field here with the value of the subscription's "custom_id" if any, but there is no
    # mention of this field in the official documentation, so I'm not sure if we should rely on it at all. The web-hook
    # processing of the integration layer does not utilize this, but queries the related captured-payment details which
    # has an official "custom_id" field.
    custom: Optional[str] = None


class WebHookEventCapturePending(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.CAPTURE.PENDING", "capture", "2.0")
    resource: WebHookCaptureResourceV2


class WebHookEventCaptureCompleted(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.CAPTURE.COMPLETED", "capture", "2.0")
    resource: WebHookCaptureResourceV2


class WebHookEventCaptureReversed(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.CAPTURE.REVERSED", "refund", "2.0")
    resource: WebHookRefundResourceV2


class WebHookEventCaptureRefunded(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.CAPTURE.REFUNDED", "refund", "2.0")
    resource: WebHookRefundResourceV2


class WebHookEventCaptureDeclined(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.CAPTURE.DECLINED", "capture", "2.0")
    resource: WebHookCaptureResourceV2


class WebHookEventSubscriptionCreated(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.CREATED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSubscriptionActivated(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.ACTIVATED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSubscriptionSuspended(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.SUSPENDED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSubscriptionExpired(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.EXPIRED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSubscriptionCancelled(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.CANCELLED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSubscriptionPaymentFailed(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.SUBSCRIPTION.PAYMENT.FAILED", "subscription", "2.0")
    resource: WebHookSubscriptionResourceV2


class WebHookEventSaleCompleted(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.SALE.COMPLETED", "sale", "1.0")
    resource: WebHookSaleResource


class WebHookEventSaleReversed(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.SALE.REVERSED", "sale", "1.0")
    resource: WebHookSaleResource


class WebHookEventSaleRefunded(WebHookEvent):
    HANDLER_KEY: ClassVar = ("PAYMENT.SALE.REFUNDED", "sale", "1.0")
    resource: WebHookSaleResource


class WebHookEventPlanCreated(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.PLAN.CREATED", "plan", "2.0")
    resource: WebHookPlanResourceV2


class WebHookEventPlanUpdated(WebHookEvent):
    HANDLER_KEY: ClassVar = ("BILLING.PLAN.UPDATED", "plan", "2.0")
    resource: WebHookPlanResourceV2



from typing import Optional, Literal, ClassVar

from pydantic import BaseModel, Field


class MonetaryValue(BaseModel):
    currency_code: str = Field(min_length=3, max_length=3)
    value: str = Field(max_length=32)


class HATEOASLink(BaseModel):
    href: str
    rel: str
    method: Optional[str] = None


class RelatedIds(BaseModel):
    order_id: Optional[str] = None
    authorization_id: Optional[str] = None
    capture_id: Optional[str] = None


class PaymentSupplementaryData(BaseModel):
    related_ids: Optional[RelatedIds] = None


class PlanPaymentPreferences(BaseModel):
    auto_bill_outstanding: Optional[bool] = True
    setup_fee_failure_action: Optional[Literal["CONTINUE", "CANCEL"]] = "CANCEL"
    payment_failure_threshold: Optional[int] = 0
    setup_fee: MonetaryValue


class PlanPricingTier(BaseModel):
    starting_quantity: str
    ending_quantity: Optional[str] = None
    amount: MonetaryValue


class PlanPricingModel(BaseModel):
    pricing_model: Optional[Literal["VOLUME", "TIERED"]] = None
    tiers: Optional[list[PlanPricingTier]] = None
    fixed_price: Optional[MonetaryValue] = None


class PlanFrequency(BaseModel):
    interval_unit: Literal["DAY", "WEEK", "MONTH", "YEAR"]
    interval_count: int = Field(ge=1, le=365, default=1)  # NOTE: maximum depends on interval_unit

    INTERVAL_DAYS: ClassVar = {"DAY": 1, "WEEK": 7, "MONTH": 31, "YEAR": 365}

    @property
    def interval_as_days(self) -> int:
        return self.INTERVAL_DAYS[self.interval_unit] * self.interval_count


class PlanBillingCycle(BaseModel):
    tenure_type: Literal["REGULAR", "TRIAL"]
    sequence: int = Field(ge=1, le=99)
    total_cycles: int = Field(ge=0, le=999, default=1)  # 0 infinite
    pricing_scheme: Optional[PlanPricingModel]  # trial needs no pricing
    frequency: PlanFrequency


class PlanTaxes(BaseModel):
    inclusive: bool = True
    percentage: str = Field(pattern=r"^((-?[0-9]+)|(-?([0-9]+)?[.][0-9]+))$")



"""
Paypal2

:copyright: 2024 Nándor Mátravölgyi
:license: Proprietary, see LICENSE for more details.
"""

__copyright__ = "Copyright 2024 Nándor Mátravölgyi"
__credits__ = ["Nándor Mátravölgyi"]
__version__ = "0.5.0"



from paypal2.client import PayPalApiClient


class PayPalDeps:
    PAYPAL: PayPalApiClient = None

    @classmethod
    async def setup_paypal(cls, *args, **kwargs):
        pp = PayPalApiClient(*args, **kwargs)
        await pp.setup()
        cls.PAYPAL = pp
        return pp

    @classmethod
    async def close_paypal(cls):
        if cls.PAYPAL:
            await cls.PAYPAL.close()

    @classmethod
    def get_paypal(cls) -> PayPalApiClient:
        pp = cls.PAYPAL
        if not pp:
            raise Exception("PAYPAL dep has not been set up")
        return pp



from datetime import datetime
from typing import Optional, Literal

from pydantic import BaseModel, Field

from paypal2.models.common import HATEOASLink, MonetaryValue, PaymentSupplementaryData


class CapturedPayment(BaseModel):
    id: str
    status: Literal["COMPLETED", "DECLINED", "PARTIALLY_REFUNDED", "PENDING", "REFUNDED", "FAILED"]
    # todo status_details
    invoice_id: Optional[str] = None
    custom_id: Optional[str] = Field(max_length=127, default=None)
    final_capture: Optional[bool] = None
    disbursement_mode: Optional[Literal["INSTANT", "DELAYED"]] = Field(
        min_length=1, max_length=16, pattern=r"^[A-Z_]+$", default="INSTANT"
    )
    amount: Optional[MonetaryValue] = None
    supplementary_data: Optional[PaymentSupplementaryData] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    links: list[HATEOASLink]
    # todo network_transaction_reference
    # todo seller_protection
    # todo seller_receivable_breakdown
    # todo payee



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from paypal2.models.common import HATEOASLink


class SubscriptionCreate(BaseModel):
    id: str = Field(min_length=6)
    quantity: Optional[str] = Field(min_length=1, max_length=32, default=None)
    custom_id: Optional[str] = Field(min_length=1, max_length=127, default=None)  # can be invoice id
    start_time: Optional[datetime] = None
    # todo shipping_amount
    # todo subscriber
    # todo application_context
    # todo plan


class SubscriptionMinimalResponse(BaseModel):
    id: str
    status: str
    links: list[HATEOASLink]


class SubscriptionDetails(BaseModel):
    id: str = Field(min_length=3, max_length=50)
    plan_id: str = Field(min_length=3, max_length=50)
    quantity: Optional[str] = Field(min_length=1, max_length=32, default=None)
    custom_id: Optional[str] = Field(min_length=1, max_length=127, default=None)
    status: Optional[str] = Field(min_length=1, max_length=24, default=None)
    status_change_note: Optional[str] = Field(min_length=1, max_length=128, default=None)
    status_update_time: Optional[str] = Field(min_length=20, max_length=64, default=None)
    plan_overridden: Optional[bool] = None
    start_time: Optional[datetime] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # todo shipping_amount
    # todo subscriber
    # todo application_context
    # todo plan


class SubscriptionReason(BaseModel):
    reason: str = Field(min_length=1, max_length=127)


class SubscriptionTransaction(BaseModel):
    id: str
    status: str
    time: str


class SubscriptionTransactionList(BaseModel):
    transactions: list[SubscriptionTransaction]
    total_items: Optional[int] = None
    total_pages: Optional[int] = None
    # todo links



from datetime import datetime
from typing import Optional, Literal

from pydantic import BaseModel, Field


class _ProductMinimalCommon(BaseModel):
    id: Optional[str] = Field(min_length=6, max_length=50, default=None)
    name: Optional[str] = Field(min_length=1, max_length=127, default=None)
    description: Optional[str] = Field(min_length=1, max_length=256, default=None)


class _ProductCommon(_ProductMinimalCommon):
    type: Optional[Literal["DIGITAL", "PHYSICAL", "SERVICE"]] = Field(min_length=1, max_length=24, default=None)
    category: Optional[str] = Field(min_length=4, max_length=256, pattern=r"^[A-Z_]+$", default=None)
    image_url: Optional[str] = Field(min_length=1, max_length=2000, default=None)
    home_url: Optional[str] = Field(min_length=1, max_length=2000, default=None)


class ProductDetails(_ProductCommon):
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # todo links


class ProductCreate(_ProductCommon):
    name: str = Field(min_length=1, max_length=127)
    type: Literal["DIGITAL", "PHYSICAL", "SERVICE"] = Field(min_length=1, max_length=24)


class ProductListProduct(_ProductMinimalCommon):
    create_time: Optional[datetime] = None
    # todo links


class ProductList(BaseModel):
    products: list[ProductListProduct]
    total_items: Optional[int] = None
    total_pages: Optional[int] = None
    # todo links



from datetime import datetime
from typing import Optional, Literal

from pydantic import BaseModel, Field

from paypal2.models.common import HATEOASLink, PlanBillingCycle, PlanPaymentPreferences, PlanTaxes


class PlanMinimalResponse(BaseModel):
    id: str
    status: Literal["CREATED", "INACTIVE", "ACTIVE"] = Field(min_length=1, max_length=24, default="ACTIVE")
    links: list[HATEOASLink]


class _PlanCommon(BaseModel):
    product_id: str = Field(min_length=6)
    name: str = Field(min_length=1, max_length=127)
    status: Literal["CREATED", "INACTIVE", "ACTIVE"] = Field(min_length=1, max_length=24, default="ACTIVE")
    description: Optional[str] = Field(min_length=1, max_length=127, default=None)
    billing_cycles: Optional[list[PlanBillingCycle]] = Field(min_length=1, max_length=12, default=None)
    payment_preferences: Optional[PlanPaymentPreferences] = None
    taxes: Optional[PlanTaxes] = None


class PlanCreate(_PlanCommon):
    payment_preferences: PlanPaymentPreferences


class PlanDetails(_PlanCommon):
    id: str = Field(min_length=6)
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # todo quantity_supported
    # todo links


class PlanListPlan(_PlanCommon):
    id: str = Field(min_length=6)
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # todo quantity_supported
    # todo links


class PlanList(BaseModel):
    plans: list[PlanListPlan]
    total_items: Optional[int] = None
    total_pages: Optional[int] = None
    # todo links


# class PlanUpdate(BaseModel):
#     op: Literal["add", "remove", "replace", "move", "copy", "test"]
#     path: str
#     value: Optional[Any] = None
#     from_field: Optional[Any] = Field(alias="from", default=None)



from paypal2.client import PayPalApiClient
from paypal2.models.order import OrderMinimalResponse, OrdersCreate


class AbstractOrderCreateProcess:
    """
    Blueprint for implementing/performing an order creation.
    """

    def __init__(self, pp: PayPalApiClient):
        self.pp = pp

    async def run(self, request: OrdersCreate) -> OrderMinimalResponse:
        await self._validate_request(request)
        try:
            order_response = await self.pp.order_create(request)
        except Exception as e:
            await self._handle_pp_exception(e)
            raise e
        await self._complete_create(order_response)
        return order_response

    async def _validate_request(self, request: OrdersCreate):
        """
        Validate the order data and any custom server-local data related to the order.
        Raise an exception if something is not correct.
        """

    async def _handle_pp_exception(self, e: Exception):
        """
        Custom handling of a PayPal API call exception can log it for review and raise a custom exception instead.
        """

    async def _complete_create(self, order_response: OrderMinimalResponse):
        """
        Store/save the id of the created order for the relevant entity, so it can be verified to match a request to
        order capturing later.
        """
        raise NotImplementedError()


class AbstractOrderCaptureProcess:
    """
    Blueprint for implementing/performing an order capture.
    """

    def __init__(self, pp: PayPalApiClient):
        self.pp = pp

    async def run(self, order_id: str) -> OrderMinimalResponse:
        """
        Data access and processing needs strict transaction properties (ACID) over the execution for safety against
        parallel multiple invocations.
        Retrying or handling differently when the "status" of the returned order data is not "COMPLETED" is up to the
        coder.
        """
        await self._validate_order_id(order_id)
        try:
            order_response = await self.pp.order_capture(order_id)
        except Exception as e:
            await self._handle_pp_exception(e)
            raise e
        if order_response.status == "COMPLETED":
            await self._complete_capture(order_response)
        return order_response

    async def _validate_order_id(self, order_id: str):
        """
        Raise an exception if the order_id is not known/saved to the relevant entities.
        """
        raise NotImplementedError()

    async def _handle_pp_exception(self, e: Exception):
        """
        Custom handling of a PayPal API call exception can log it for review and raise a custom exception instead.
        """

    async def _complete_capture(self, order_response: OrderMinimalResponse):
        """
        Honor the captured order by granting the purchased service/thing to the user.
        Also clear the stored order_response.id so a subsequent call to _validate_order_id() will not succeed.
        """
        raise NotImplementedError()



from datetime import datetime
from typing import Optional, Literal

from pydantic import BaseModel, Field

from paypal2.models.common import MonetaryValue, HATEOASLink


class Breakdown(BaseModel):
    item_total: Optional[MonetaryValue] = None
    shipping: Optional[MonetaryValue] = None
    handling: Optional[MonetaryValue] = None
    tax_total: Optional[MonetaryValue] = None
    insurance: Optional[MonetaryValue] = None
    shipping_discount: Optional[MonetaryValue] = None
    discount: Optional[MonetaryValue] = None


class MonetaryValueWithBreakdown(MonetaryValue):
    breakdown: Optional[Breakdown] = None


class Item(BaseModel):
    name: str = Field(min_length=1, max_length=127)
    quantity: str = Field(min_length=1, max_length=10)
    description: Optional[str] = Field(min_length=1, max_length=127, default=None)
    sku: Optional[str] = Field(min_length=1, max_length=127, default=None)
    url: Optional[str] = Field(min_length=1, max_length=2048, default=None)
    category: Optional[Literal["DIGITAL_GOODS", "PHYSICAL_GOODS", "DONATION"]] = Field(
        min_length=1, max_length=20, default=None
    )
    image_url: Optional[str] = Field(
        min_length=1,
        max_length=2048,
        pattern=r"^(https:)([/|.|\w|\s|-])*\.(?:jpg|gif|png|jpeg|JPG|GIF|PNG|JPEG)",
        default=None,
    )
    unit_amount: MonetaryValue  # purchase_units[].amount.breakdown.item_total must be set if present
    tax: Optional[MonetaryValue] = None  # purchase_units[].amount.breakdown.tax_total must be set if present
    # TODO: upc


class PurchaseUnit(BaseModel):
    reference_id: Optional[str] = Field(min_length=1, max_length=256, default=None)
    description: Optional[str] = Field(min_length=1, max_length=127, default=None)
    custom_id: Optional[str] = Field(min_length=1, max_length=255, default=None)
    invoice_id: Optional[str] = Field(min_length=1, max_length=127, default=None)
    soft_descriptor: Optional[str] = Field(min_length=1, max_length=22, default=None)
    items: Optional[list[Item]] = None
    amount: MonetaryValueWithBreakdown
    # TODO: payee
    # TODO: payment_instruction
    # TODO: shipping
    # TODO: supplementary_data


class OrdersCreate(BaseModel):
    purchase_units: list[PurchaseUnit] = Field(min_length=1, max_length=10)
    intent: Optional[Literal["CAPTURE", "AUTHORIZE"]]
    # TODO: payment_source


class OrderMinimalResponse(BaseModel):
    id: str
    status: str
    links: list[HATEOASLink]


class OrderCreated(OrderMinimalResponse):
    processing_instruction: Optional[Literal["ORDER_COMPLETE_ON_PAYMENT_APPROVAL", "NO_INSTRUCTION"]] = Field(
        min_length=1, max_length=36, pattern=r"^[0-9A-Z_]+$", default="NO_INSTRUCTION"
    )
    intent: Optional[Literal["CAPTURE", "AUTHORIZE"]] = None
    create_time: Optional[datetime] = None
    update_time: Optional[datetime] = None
    # todo payer
    # todo purchase_units



import base64
import json
import ssl
import zlib
from contextlib import asynccontextmanager
from datetime import datetime, timezone, timedelta
from typing import Any
from typing import Optional, Literal

import aiohttp
import certifi
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
from aiohttp import TCPConnector
from pydantic import BaseModel
from ttldict2 import TTLDict

from paypal2.models.order import OrderMinimalResponse
from paypal2.models.order import OrdersCreate
from paypal2.models.payment import CapturedPayment
from paypal2.models.plan import PlanCreate, PlanDetails, PlanList, PlanMinimalResponse
from paypal2.models.product import ProductDetails, ProductCreate, ProductList
from paypal2.models.subscription import (
    SubscriptionMinimalResponse,
    SubscriptionCreate,
    SubscriptionDetails,
    SubscriptionReason,
    SubscriptionTransactionList,
)


class AccessTokenReset(Exception):
    pass


class JsonBadRequest(aiohttp.ClientResponseError):
    @classmethod
    def from_cre(cls, cre: aiohttp.ClientResponseError, data: Any):
        return cls(
            cre.request_info,
            cre.history,
            code=cre.code,
            status=cre.status,
            message=cre.message,
            headers=cre.headers,
            data=data,
        )

    def __init__(self, request_info, history, *, code=None, status=None, message="", headers=None, data=None) -> None:
        aiohttp.ClientResponseError.__init__(
            self, request_info, history, code=code, status=status, message=message, headers=headers
        )
        self.data = data


class AbstractExternalWebHookCertificateStore:
    async def get(self, url: str) -> Optional[bytes]:
        raise NotImplementedError()

    async def set(self, url: str, content: bytes):
        raise NotImplementedError()


class PayPalApiClient:
    __slots__ = (
        "client_id",
        "__client_secret",
        "url_base",
        "proxy_address",
        "session",
        "session_owned",
        "webhook_id",
        "external_cert_store",
        "_access_token",
        "_cert_cache",
    )

    PRODUCTION_URL_BASE = "https://api-m.paypal.com"
    SANDBOX_URL_BASE = "https://api-m.sandbox.paypal.com"

    @classmethod
    def init_session(cls):
        return aiohttp.ClientSession(
            connector=TCPConnector(ssl=ssl.create_default_context(cafile=certifi.where()), limit=16)
        )

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        url_base: str = None,
        proxy_address: str = None,
        session: aiohttp.ClientSession = None,
        webhook_id: str = None,
        external_cert_store: Optional[AbstractExternalWebHookCertificateStore] = None,
    ):
        self.client_id = client_id
        self.__client_secret = client_secret
        self.url_base = (url_base or self.PRODUCTION_URL_BASE).rstrip("/")
        self.proxy_address = proxy_address
        self.session_owned = session is None
        self.session = session
        self.webhook_id = webhook_id
        self.external_cert_store = external_cert_store
        # state
        self._access_token = None
        self._cert_cache = TTLDict(14 * 24 * 60 * 60.0, max_items=10)

    async def setup(self, session: aiohttp.ClientSession = None):
        if self.session is None:
            self.session_owned = session is None
            if self.session_owned:
                session = self.init_session()
            self.session = session

    async def close(self):
        if self.session is not None:
            if self.session_owned:
                await self.session.close()
            self.session = None

    async def _get_access_token(self):
        token = base64.b64encode(f"{self.client_id}:{self.__client_secret}".encode()).decode()
        async with self.session.post(
            self.url_base + "/v1/oauth2/token",
            data="grant_type=client_credentials",
            headers={"Content-Type": "application/x-www-form-urlencoded", "Authorization": f"Basic {token}"},
            proxy=self.proxy_address,
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()
        access_token = data.get("access_token")
        if not isinstance(access_token, str):
            raise Exception("PayPal auth error")
        return access_token

    @asynccontextmanager
    async def _manage_access_token(self, raise_reset: bool = False):
        if not self._access_token:
            self._access_token = await self._get_access_token()
        access_token = self._access_token
        try:
            yield access_token
        except aiohttp.ClientResponseError as e:
            if e.status == 401:
                if self._access_token == access_token:
                    self._access_token = None
                if raise_reset:
                    raise AccessTokenReset() from e
            raise e

    async def _retry_request(
        self,
        method: Literal["get"] | Literal["post"] | Literal["patch"],
        url: str,
        body: BaseModel | dict | list | None = None,
        headers: dict[str, str] = None,
        params: Optional[dict[str, str]] = None,
        retry: int = 2,
        response_body: bool = True,
    ):
        j = body.model_dump(mode="json", exclude_unset=True) if isinstance(body, BaseModel) else body
        for i in range(1, retry + 1):
            try:
                async with self._manage_access_token(raise_reset=i < retry) as access_token:
                    hs = {"Authorization": f"Bearer {access_token}"}
                    if headers:
                        hs.update(headers)
                    async with getattr(self.session, method)(
                        self.url_base + url, json=j, headers=hs, params=params, proxy=self.proxy_address
                    ) as resp:
                        try:
                            resp.raise_for_status()
                        except aiohttp.ClientResponseError as e:
                            if e.status == 400:
                                try:
                                    details = await resp.json()
                                except Exception:
                                    raise e
                                else:
                                    raise JsonBadRequest.from_cre(e, details)
                            raise e
                        if response_body:
                            return await resp.json()
            except AccessTokenReset:
                pass

    async def order_create(self, request: OrdersCreate) -> OrderMinimalResponse:
        """
        Create an order for purchase approval.
        """
        data = await self._retry_request("post", "/v2/checkout/orders", request)
        return OrderMinimalResponse.model_validate(data)

    async def order_capture(self, order_id: str) -> OrderMinimalResponse:
        """
        :param order_id: Must be checked that this service created the order previously, and knows how to handle the
            capture if it is completed.
        :return: the status property indicates success if "COMPLETED"
        """
        data = await self._retry_request(
            "post", f"/v2/checkout/orders/{order_id}/capture", {}, {"Prefer": "return=minimal"}
        )
        return OrderMinimalResponse.model_validate(data)

    async def subscription_create(self, request: SubscriptionCreate) -> SubscriptionMinimalResponse:
        """
        Create a subscription for client approval.
        """
        data = await self._retry_request("post", "/v1/billing/subscriptions", request)
        return SubscriptionMinimalResponse.model_validate(data)

    async def subscription_details(self, subscription_id: str) -> Optional[SubscriptionDetails]:
        try:
            data = await self._retry_request("get", f"/v1/billing/subscriptions/{subscription_id}")
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise e
        return SubscriptionDetails.model_validate(data)

    async def subscription_activate(self, subscription_id: str, request: SubscriptionReason):
        await self._retry_request(
            "post", f"/v1/billing/subscriptions/{subscription_id}/activate", request, response_body=False
        )

    async def subscription_suspend(self, subscription_id: str, request: SubscriptionReason):
        await self._retry_request(
            "post", f"/v1/billing/subscriptions/{subscription_id}/suspend", request, response_body=False
        )

    async def subscription_cancel(self, subscription_id: str, request: SubscriptionReason):
        await self._retry_request(
            "post", f"/v1/billing/subscriptions/{subscription_id}/cancel", request, response_body=False
        )

    async def subscription_transaction_list(
        self, subscription_id: str, start_time: datetime = None, end_time: datetime = None
    ) -> Optional[SubscriptionTransactionList]:
        if not start_time:
            start_time = datetime.utcnow() - timedelta(minutes=60)
        if not end_time:
            end_time = datetime.utcnow() + timedelta(minutes=10)
        try:
            data = await self._retry_request(
                "get",
                f"/v1/billing/subscriptions/{subscription_id}/transactions",
                params={
                    "start_time": start_time.replace(tzinfo=timezone.utc).isoformat(timespec="seconds"),
                    "end_time": end_time.replace(tzinfo=timezone.utc).isoformat(timespec="seconds"),
                },
            )
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise e
        return SubscriptionTransactionList.model_validate(data)

    async def product_create(self, request: ProductCreate) -> ProductDetails:
        data = await self._retry_request("post", "/v1/catalogs/products", request, {"Prefer": "return=representation"})
        return ProductDetails.model_validate(data)

    async def product_list(self) -> ProductList:
        data = await self._retry_request("get", "/v1/catalogs/products")
        return ProductList.model_validate(data)

    async def product_details(self, product_id: str) -> Optional[ProductDetails]:
        try:
            data = await self._retry_request("get", f"/v1/catalogs/products/{product_id}")
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise e
        return ProductDetails.model_validate(data)

    async def plan_create(self, request: PlanCreate) -> PlanMinimalResponse:
        data = await self._retry_request("post", "/v1/billing/plans", request)
        return PlanMinimalResponse.model_validate(data)

    async def plan_list(self, product_id: str = None, page_size: int = None, page: int = None) -> PlanList:
        params = {}
        if product_id:
            params["product_id"] = product_id
        if page_size:
            params["page_size"] = str(page_size)
        if page:
            params["page"] = str(page)
        data = await self._retry_request(
            "get", "/v1/billing/plans", None, {"Prefer": "return=representation"}, params or None
        )
        return PlanList.model_validate(data)

    async def plan_details(self, plan_id: str) -> Optional[PlanDetails]:
        try:
            data = await self._retry_request("get", f"/v1/billing/plans/{plan_id}")
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise e
        return PlanDetails.model_validate(data)

    # async def plan_update(self, plan_id: str, ops: list[PlanUpdate]):
    #     await self._retry_request("patch", f"/v1/billings/plans/{plan_id}", ops, response_body=False)

    async def captured_payment_details(self, capture_id: str) -> Optional[CapturedPayment]:
        try:
            data = await self._retry_request("get", f"/v2/payments/captures/{capture_id}")
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise e
        return CapturedPayment.model_validate(data)

    async def verify_webhook_notification(self, body: bytes, headers: dict[str, str], webhook_id: str = None) -> dict:
        """
        WebHook notification verification.
        """
        webhook_id = webhook_id or self.webhook_id
        if not isinstance(webhook_id, str):
            raise ValueError("Paypal webhook id not configured")
        transmission_auth_algo = headers.get("paypal-auth-algo", "SHA256withRSA")  # default
        if transmission_auth_algo != "SHA256withRSA":
            raise ValueError(f"Paypal webhook signature algorithm not supported: {transmission_auth_algo}")
        transmission_signature = headers.get("paypal-transmission-sig")
        transmission_id = headers.get("paypal-transmission-id")
        transmission_time = headers.get("paypal-transmission-time")
        transmission_cert_url = headers.get("paypal-cert-url")
        if (
            not isinstance(transmission_signature, str)
            or not isinstance(transmission_id, str)
            or not isinstance(transmission_time, str)
            or not isinstance(transmission_cert_url, str)
        ):
            raise ValueError("Paypal webhook headers invalid")
        if self.url_base != self.SANDBOX_URL_BASE and not transmission_cert_url.startswith("https://api.paypal.com/"):
            raise ValueError("Paypal webhook cert url is not from PayPal API")
        rsa_pub_key = self._cert_cache.get(transmission_cert_url, touch=True)
        if not rsa_pub_key:
            transmission_cert = None
            if self.external_cert_store:
                transmission_cert = await self.external_cert_store.get(transmission_cert_url)
            if not transmission_cert:
                async with self.session.get(transmission_cert_url) as r:
                    r.raise_for_status()
                    transmission_cert = await r.read()
            rsa_pub_key = RSA.import_key(transmission_cert)
            if rsa_pub_key.has_private():
                raise ValueError("Paypal webhook transmission key imported as private")
            if self.external_cert_store:
                await self.external_cert_store.set(transmission_cert_url, transmission_cert)
            self._cert_cache[transmission_cert_url] = rsa_pub_key
        # TODO: crc, hash, signature check and json load could be offloaded to worker thread?
        body_checksum = zlib.crc32(body)
        message = f"{transmission_id}|{transmission_time}|{webhook_id}|{body_checksum}"
        hashed = SHA256.new(message.encode())
        pkcs1_15.new(rsa_pub_key).verify(hashed, base64.b64decode(transmission_signature))
        data = json.loads(body)
        if not isinstance(data, dict):
            raise ValueError("Paypal webhook data is not dict")
        return data



import os
import shutil
from collections import defaultdict

from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))
VERSION = "0.5.0"

with open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = defaultdict(list)
for name in os.listdir(os.path.join(HERE, "requirements")):
    if name not in ("base.in", "test.in"):
        continue
    reqs = requirements[name.rpartition(".")[0]]
    with open(os.path.join(HERE, "requirements", name)) as f:
        for line in f:
            line = line.strip()
            if line.startswith("#") or line.startswith("-r"):
                continue
            reqs.append(line)
install_requirements = requirements.pop("base")


# Manually cleaning before build is required.
for p in [os.path.join(HERE, "build"), os.path.join(HERE, "dist"), os.path.join(HERE, "paypal2.egg-info")]:
    if os.path.exists(p):
        shutil.rmtree(p)

setup(
    name="paypal2",
    version=VERSION,
    description="Opinionated asyncio Paypal server-side integration SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Traktormaster/paypal2",
    author="Nándor Mátravölgyi",
    author_email="nandor.matra@gmail.com",
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=[
        p for p in find_packages(where=HERE, include=["paypal2*"]) if not p.startswith("paypal2_")
    ],
    # package_data=package_data,
    install_requires=install_requirements,
    # entry_points={"console_scripts": console_scripts},
    extras_require=dict(requirements),
    python_requires=">=3.10",
)
